#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that adds interactions from blocks to molecules.
"""
# TODO: Move all this functionality to do_mapping?
from collections import ChainMap
from itertools import product

from .processor import Processor
from ..graph_utils import make_residue_graph
from ..molecule import Molecule


def apply_blocks(molecule, blocks):
    """
    Generate a new :class:`~vermouth.molecule.Molecule` based on the residue
    names and other attributes of `molecule` from `blocks`.

    Parameters
    ----------
    molecule: vermouth.molecule.Molecule
        The molecule to process.
    blocks: dict[str, vermouth.molecule.Block]
        The blocks known.

    Returns
    -------
    vermouth.molecule.Molecule
        A new molecule with attributes from the old `molecule`, as well as all
        interactions described by `blocks`.
    """
    graph_out = Molecule(
        force_field=molecule.force_field,
        meta=molecule.meta.copy()
    )
    residue_graph = make_residue_graph(molecule)

    # nrexcl may not be defined, but if it is we probably want to keep it
    try:
        graph_out.nrexcl = molecule.nrexcl
    except AttributeError:
        graph_out.nrexcl = None

    old_to_new_idxs = {}
    at_idx = 0
    charge_group_offset = 0
    for res_idx in residue_graph:
        residue = residue_graph.nodes[res_idx]
        res_graph = residue['graph']
        resname = residue['resname']
        block = blocks[resname]
        atname_to_idx = {}

        if graph_out.nrexcl is None:
            if hasattr(block, 'nrexcl'):
                graph_out.nrexcl = block.nrexcl
        else:
            if (hasattr(block, 'nrexcl')
                    and block.nrexcl is not None
                    and block.nrexcl != graph_out.nrexcl):
                raise ValueError('Not all blocks share the same value for "nrexcl".')

        for block_idx in block:
            atname = block.nodes[block_idx]['atomname']
            atom = list(res_graph.find_atoms(atomname=atname))
            assert len(atom) == 1, (block.name, atname, atom)
            old_to_new_idxs[atom[0]] = at_idx
            atname_to_idx[atname] = at_idx
            attrs = molecule.nodes[atom[0]]
            graph_out.add_node(at_idx, **ChainMap(block.nodes[atname], attrs))
            graph_out.nodes[at_idx]['graph'] = molecule.subgraph(atom)
            graph_out.nodes[at_idx]['charge_group'] += charge_group_offset
            graph_out.nodes[at_idx]['resid'] = attrs['resid']
            at_idx += 1
        charge_group_offset = graph_out.nodes[at_idx - 1]['charge_group']
        for idx, jdx, data in block.edges(data=True):
            idx = atname_to_idx[idx]
            jdx = atname_to_idx[jdx]
            graph_out.add_edge(idx, jdx, **data)
        for inter_type, interactions in block.interactions.items():
            for interaction in interactions:
                atom_idxs = []
                for atom_name in interaction.atoms:
                    atom_index = graph_out.find_atoms(atomname=atom_name,
                                                      resname=residue['resname'],
                                                      resid=residue['resid'])
                    atom_index = list(atom_index)
                    if not atom_index:
                        msg = ('Could not find a atom named "{}" '
                               'with resname being "{}" '
                               'and resid being "{}".')
                        raise ValueError(msg.format(atom_name, residue['resname'], residue['resid']))
                    atom_idxs.extend(atom_index)
                interactions = interaction._replace(atoms=atom_idxs)
                graph_out.add_interaction(inter_type, *interactions)

    # This makes edges between residues. We need to do this, since they can't
    # come from the blocks and we need them to find the links locations.
    # TODO This should not be done here, but by do_mapping, which might *also*
    #      do it at the moment
    for res_idx, res_jdx in residue_graph.edges():
        for old_idx, old_jdx in product(residue_graph.nodes[res_idx]['graph'],
                                        residue_graph.nodes[res_jdx]['graph']):
            try:
                # Usually termini, PTMs, etc
                idx = old_to_new_idxs[old_idx]
                jdx = old_to_new_idxs[old_jdx]
            except KeyError:
                continue
            if molecule.has_edge(old_idx, old_jdx):
                graph_out.add_edge(idx, jdx)
    return graph_out



class ApplyBlocks(Processor):
    def run_molecule(self, molecule):
        return apply_blocks(molecule, molecule.force_field.blocks)



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides functions for reading and writing PDB files.
"""

from functools import partial

import numpy as np

from ..molecule import Molecule
from ..utils import first_alpha, distance
from ..truncating_formatter import TruncFormatter


def get_not_none(node, attr, default):
    """
    Returns ``node[attr]``. If it doesn't exists or is ``None``, return
    `default`.

    Parameters
    ----------
    node: collections.abc.Mapping
    attr: collections.abc.Hashable
    default
        The value to return if ``node[attr]`` is either ``None``, or does not
        exist.

    Returns
    -------
    object
        The value of ``node[attr]`` if it exists and is not ``None``, else
        `default`.
    """
    value = node.get(attr)
    if value is None:
        value = default
    return value


def write_pdb_string(system, conect=True, omit_charges=True, nan_missing_pos=False):
    """
    Describes `system` as a PDB formatted string. Will create CONECT records
    from the edges in the molecules in `system` iff `conect` is True.

    Parameters
    ----------
    system: vermouth.system.System
        The system to write.
    conect: bool
        Whether to write CONECT records for the edges.
    omit_charges: bool
        Whether charges should be omitted. This is usually a good idea since
        the PDB format can only deal with integer charges.
    nan_missing_pos: bool
        Wether the writing should fail if an atom does not have a position.
        When set to `True`, atoms without coordinates will be written
        with 'nan' as coordinates; this will cause the output file to be
        *invalid* for most uses.
        for most use.

    Returns
    -------
    str
        The system as PDB formatted string.
    """
    out = []

    formatter = TruncFormatter()
#    format_string = 'ATOM  {: >5.5d} {:4.4s}{:1.1s}{:3.3s} {:1.1s}{:4.4d}{:1.1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:2.2s}{:2.2s}'
    format_string = 'ATOM  {: >5dt} {:4st}{:1st}{:3st} {:1st}{:>4dt}{:1st}   {:8.3ft}{:8.3ft}{:8.3ft}{:6.2ft}{:6.2ft}          {:2st}{:2st}'

    # FIXME Here we make the assumption that node indices are unique across
    # molecules in a system. Probably not a good idea
    nodeidx2atomid = {}
    atomid = 1
    for mol_idx, molecule in enumerate(system.molecules):
        node_order = molecule.nodes

        for node_idx in node_order:
            nodeidx2atomid[(mol_idx, node_idx)] = atomid
            node = molecule.node[node_idx]
            atomname = node['atomname']
            altloc = get_not_none(node, 'altloc', '')
            resname = node['resname']
            chain = node['chain']
            resid = node['resid']
            insertion_code = get_not_none(node, 'insertioncode', '')
            try:
                # converting from nm to A
                x, y, z = node['position'] * 10  # pylint: disable=invalid-name
            except KeyError:
                if nan_missing_pos:
                    x = y = z = float('nan')  # pylint: disable=invalid-name
                else:
                    raise
            occupancy = get_not_none(node, 'occupancy', 1)
            temp_factor = get_not_none(node, 'temp_factor', 0)
            element = get_not_none(node, 'element', '')
            charge = get_not_none(node, 'charge', 0)
            if charge and not omit_charges:
                charge = '{:+2d}'.format(int(charge))[::-1]
            else:
                charge = ''
            line = formatter.format(format_string, atomid, atomname, altloc,
                                    resname, chain, resid, insertion_code, x,
                                    y, z, occupancy, temp_factor, element,
                                    charge)
            atomid += 1
            out.append(line)
        terline = formatter.format('TER   {: >5dt}      {:3st} {:1st}{: >4dt}{:1st}',
                                   atomid, resname, chain, resid, insertion_code)
        atomid += 1
        out.append(terline)
    if conect:
        number_fmt = '{:>4dt}'
        format_string = 'CONECT '
        for mol_idx, molecule in enumerate(system.molecules):
            node_order = molecule.nodes

            for node_idx in node_order:
                todo = [nodeidx2atomid[(mol_idx, n_idx)]
                        for n_idx in molecule[node_idx] if n_idx > node_idx]
                while todo:
                    current, todo = todo[:4], todo[4:]
                    fmt = ['CONECT'] + [number_fmt]*(len(current) + 1)
                    fmt = ' '.join(fmt)
                    line = formatter.format(fmt, nodeidx2atomid[(mol_idx, node_idx)], *current)
                    out.append(line)
    out.append('END   ')
    return '\n'.join(out)


def write_pdb(system, path, conect=True, omit_charges=True, nan_missing_pos=False):
    """
    Writes `system` to `path` as a PDB formatted string.

    Parameters
    ----------
    system: vermouth.system.System
        The system to write.
    path: str
        The file to write to.
    conect: bool
        Whether to write CONECT records for the edges.
    omit_charges: bool
        Whether charges should be omitted. This is usually a good idea since
        the PDB format can only deal with integer charges.
    nan_missing_pos: bool
        Wether the writing should fail if an atom does not have a position.
        When set to `True`, atoms without coordinates will be written
        with 'nan' as coordinates; this will cause the output file to be
        *invalid* for most uses.
        for most use.

    See Also
    --------
    :func:write_pdb_string
    """
    with open(path, 'w') as out:
        out.write(write_pdb_string(system, conect, omit_charges, nan_missing_pos))


def do_conect(mol, conectlist):
    """Apply connections to molecule based on CONECT records read from PDB file

    Parameters
    ----------
    mol: networkx.Graph
        The graph to add edges to.
    conectlist: collections.abc.Iterable[str]
        An iterable of CONECT records as found in a PDB file.
    """
    atidx2nodeidx = {node_data['atomid']: node_idx
                     for node_idx, node_data in mol.node.items()}

    for line in conectlist:
        start = 6
        width = 5
        ats = []
        for num in range(start, len(line.rstrip()), width):
            atom = int(line[num:num + width])
            ats.append(atom)
            try:
                at0 = atidx2nodeidx[ats[0]]
            except KeyError:
                continue
            for atom in ats[1:]:
                try:
                    atom = atidx2nodeidx[atom]
                except KeyError:
                    continue
                dist = distance(mol.node[at0]['position'], mol.node[atom]['position'])
                mol.add_edge(at0, atom, distance=dist)


def read_pdb(file_name, exclude=('SOL',), ignh=False, model=0):
    """
    Parse a PDB file to create a molecule.

    Parameters
    ----------
    filename: str
        The file to read.
    exclude: collections.abc.Container[str]
        Atoms that have one of these residue names will not be included.
    ignh: bool
        Whether hydrogen atoms should be ignored.
    model: int
        If the PDB file contains multiple models, which one to select.

    Returns
    -------
    vermouth.molecule.Molecule
        The parsed molecules. Will only contain edges if the PDB file has
        CONECT records. Either way, might be disconnected.
    """
    models = [Molecule()]
    conect = []
    idx = 0

    field_widths = (-6, 5, -1, 4, 1, 4, 1, 4, 1, -3, 8, 8, 8, 6, 6, -10, 2, 2)
    field_types = (int, str, str, str, str, int, str, float, float, float, float, float, str, str)
    field_names = ('atomid', 'atomname', 'altloc', 'resname', 'chain', 'resid',
                   'insertion_code', 'x', 'y', 'z', 'occupancy', 'temp_factor',
                   'element', 'charge')

    start = 0
    slices = []
    for width in field_widths:
        if width > 0:
            slices.append(slice(start, start + width))
        start = start + abs(width)

    with open(str(file_name)) as pdb:
        for line in pdb:
            record = line[:6]
            if record == 'ENDMDL':
                models.append(Molecule())
            elif record in ('ATOM  ', 'HETATM'):
                properties = {}
                for name, type_, slice_ in zip(field_names, field_types, slices):
                    properties[name] = type_(line[slice_].strip())

                pos = (properties.pop('x'), properties.pop('y'), properties.pop('z'))
                # Coordinates are read in Angstrom, but we want them in nm
                properties['position'] = np.array(pos, dtype=float) / 10

                if not properties['element']:
                    atomname = properties['atomname']
                    properties['element'] = first_alpha(atomname)
                if properties['resname'] in exclude or (ignh and properties['element'] == 'H'):
                    continue
                models[-1].add_node(idx, **properties)
                idx += 1
            elif record == 'CONECT':
                conect.append(line)

    if not models[-1]:
        models.pop()

    for molecule in models:
        do_conect(molecule, conect)

    molecule = models[model]

#    if molecule.number_of_edges() == 0:
#        edges_from_distance(molecule)
#        # Make all edges based on threshold distance
#        positions = np.array([molecule.node[n]['position'] for n in molecule])
#        distances = ssd.squareform(ssd.pdist([molecule.node[n]['position'] for n in molecule]))
#        idxs = np.where((distances < threshold) & (distances != 0))
#        weights = 1/distances[idxs]
#        molecule.add_weighted_edges_from(zip(idxs[0], idxs[1], weights))

    return molecule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

from vermouth.processors.do_mapping import do_mapping
import vermouth.forcefield
from vermouth.molecule import Molecule, Block
import networkx.algorithms.isomorphism as iso


FF_MARTINI = vermouth.forcefield.get_native_force_field(name='martini22')
FF_UNIVERSAL = vermouth.forcefield.get_native_force_field(name='universal')

AA_MOL = Molecule(force_field=FF_UNIVERSAL)
AA_MOL.add_nodes_from((
        (0, {'resid': 1, 'resname': 'IPO', 'atomname': 'C1', 'chain': 'A', 'element': 'C'}),
        (1, {'resid': 1, 'resname': 'IPO', 'atomname': 'C2', 'chain': 'A', 'element': 'C'}),
        (2, {'resid': 1, 'resname': 'IPO', 'atomname': 'C3', 'chain': 'A', 'element': 'C'}),
        (3, {'resid': 2, 'resname': 'IPO', 'atomname': 'C1', 'chain': 'A', 'element': 'C'}),
        (4, {'resid': 2, 'resname': 'IPO', 'atomname': 'C2', 'chain': 'A', 'element': 'C'}),
        (5, {'resid': 2, 'resname': 'IPO', 'atomname': 'C3', 'chain': 'A', 'element': 'C'}),
        (6, {'resid': 3, 'resname': 'IPO', 'atomname': 'C1', 'chain': 'A', 'element': 'C'}),
        (7, {'resid': 3, 'resname': 'IPO', 'atomname': 'C2', 'chain': 'A', 'element': 'C'}),
        (8, {'resid': 3, 'resname': 'IPO', 'atomname': 'C3', 'chain': 'A', 'element': 'C'}),
        ))
AA_MOL.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4), (4, 5), (4, 6), (6, 7), (7, 8)])
block_aa = Block(force_field=FF_UNIVERSAL)
block_aa.name = 'IPO'
block_aa.add_nodes_from((
    ('C1', {'resid': 1, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C2', {'resid': 1, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C3', {'resid': 1, 'resname': 'IPO', 'atomname': 'C3'}),))
block_aa.add_edges_from([('C1', 'C2'), ('C2', 'C3')])

block_cg = Block(force_field=FF_MARTINI)
block_cg.name = 'IPO'
block_cg.add_nodes_from((('B1', {'resid': 1, 'resname': 'IPO', 'atomname': 'B1'}),))

FF_MARTINI.blocks['IPO'] = block_cg
FF_UNIVERSAL.blocks['IPO'] = block_aa


block_aa = Block(force_field=FF_UNIVERSAL)
block_aa.name = 'IPO_large'
block_aa.add_nodes_from((
    ('C1', {'resid': 1, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C2', {'resid': 1, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C3', {'resid': 1, 'resname': 'IPO', 'atomname': 'C3'}),
    ('C4', {'resid': 2, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C5', {'resid': 2, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C6', {'resid': 2, 'resname': 'IPO', 'atomname': 'C3'}),
    ('C7', {'resid': 3, 'resname': 'IPO', 'atomname': 'C1'}),
    ('C8', {'resid': 3, 'resname': 'IPO', 'atomname': 'C2'}),
    ('C9', {'resid': 3, 'resname': 'IPO', 'atomname': 'C3'}),))
block_aa.add_edges_from([('C1', 'C2'), ('C2', 'C3'), ('C2', 'C4'),
                         ('C4', 'C5'), ('C5', 'C6'), ('C5', 'C7'),
                         ('C7', 'C8'), ('C8', 'C9'),
                         ])

block_cg = Block(force_field=FF_MARTINI)
block_cg.name = 'IPO_large'
block_cg.add_nodes_from((('B1', {'resid': 1, 'resname': 'IPO_large', 'atomname': 'B1'}),
                         ('B2', {'resid': 2, 'resname': 'IPO_large', 'atomname': 'B1'}),
                         ('B3', {'resid': 3, 'resname': 'IPO_large', 'atomname': 'B1'}),))
block_cg.add_edges_from([('B1', 'B2'), ('B2', 'B3')])

FF_MARTINI.blocks['IPO_large'] = block_cg
FF_UNIVERSAL.blocks['IPO_large'] = block_aa


def _equal_graphs(g1, g2):
    attrs = ['resid', 'resname', 'atomname', 'chain', 'charge_group', 'atype']
    node_equal = iso.categorical_node_match(attrs, ['']*len(attrs))
    matcher = iso.GraphMatcher(g1, g2, node_match=node_equal)
    return matcher.is_isomorphic()


def test_no_residue_crossing():
    """
    Make sure we don't cross residue boundaries
    """
    mapping = {(0, 'C1'): [(0, 'B1')], (0, 'C2'): [(0, 'B1')], (0, 'C3'): [(0, 'B1')]}
    weights = {(0, 'B1'): {(0, 'C1'): 1, (0, 'C2'): 1, (0, 'C3'): 1, }}
    extra = ()
    mappings = {'universal': {'martini22': {'IPO': (mapping, weights, extra)}}}

    cg = do_mapping(AA_MOL, mappings, FF_MARTINI)

    expected = Molecule(force_field=FF_MARTINI)
    expected.add_nodes_from((
        (0, {'resid': 1, 'resname': 'IPO', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
        (1, {'resid': 2, 'resname': 'IPO', 'atomname': 'B1', 'chain': 'A', 'charge_group': 2}),
        (2, {'resid': 3, 'resname': 'IPO', 'atomname': 'B1', 'chain': 'A', 'charge_group': 3}),
            ))
    expected.add_edges_from(([0, 1], [1, 2]))
    
    print(cg.nodes(data=True))
    print(cg.edges())
    print('-'*80)
    print(expected.nodes(data=True))
    print(expected.edges())
    assert _equal_graphs(cg, expected)


def test_residue_crossing():
    '''
    Make sure we do cross residue boundaries and can rename residues
    '''
    mapping = {(0, 'C1'): [(0, 'B1')], (0, 'C2'): [(0, 'B1')], (0, 'C3'): [(0, 'B1')],
               (1, 'C1'): [(1, 'B1')], (1, 'C2'): [(1, 'B1')], (1, 'C3'): [(1, 'B1')],
               (2, 'C1'): [(2, 'B1')], (2, 'C2'): [(2, 'B1')], (2, 'C3'): [(2, 'B1')]}
    weights = {(0, 'B1'): {(0, 'C1'): 1, (0, 'C2'): 1, (0, 'C3'): 1, },
               (1, 'B1'): {(1, 'C1'): 1, (1, 'C2'): 1, (1, 'C3'): 1, },
               (2, 'B1'): {(2, 'C1'): 1, (2, 'C2'): 1, (2, 'C3'): 1, },}
    extra = ()
    mappings = {'universal': {'martini22': {'IPO_large': (mapping, weights, extra)}}}

    cg = do_mapping(AA_MOL, mappings, FF_MARTINI)

    expected = Molecule()
    expected.add_nodes_from((
        (0, {'resid': 1, 'resname': 'IPO_large', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
        (1, {'resid': 2, 'resname': 'IPO_large', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
        (2, {'resid': 3, 'resname': 'IPO_large', 'atomname': 'B1', 'chain': 'A', 'charge_group': 1}),
            ))
    expected.add_edges_from(([0, 1], [1, 2]))
    
    print(cg.nodes(data=True))
    print(cg.edges())
    print('-'*80)
    print(expected.nodes(data=True))
    print(expected.edges())
    assert _equal_graphs(cg, expected)


def _map_weights(mapping):
    inv_map = defaultdict(list)
    for from_, tos in mapping.items():
        for to in tos:
            inv_map[to].append(from_)
    weights = {}
    for bd, atoms in inv_map.items():
        weights[bd] = {atom: 1 for atom in atoms}
    return weights


def test_peptide():
    """
    Test multiple cg beads in residue
    """
    gly = {(0, 'C'): [(0, 'BB')], (0, 'N'): [(0, 'BB')], (0, 'O'): [(0, 'BB')], (0, 'CA'): [(0, 'BB')]}
    ile = {(0, 'C'): [(0, 'BB')], (0, 'N'): [(0, 'BB')], (0, 'O'): [(0, 'BB')], (0, 'CA'): [(0, 'BB')],
           (0, 'CB'): [(0, 'SC1')], (0, 'CG1'): [(0, 'SC1')], (0, 'CG2'): [(0, 'SC1')], (0, 'CD'): [(0, 'SC1')]}
    leu = {(0, 'C'): [(0, 'BB')], (0, 'N'): [(0, 'BB')], (0, 'O'): [(0, 'BB')], (0, 'CA'): [(0, 'BB')],
           (0, 'CB'): [(0, 'SC1')], (0, 'CG1'): [(0, 'SC1')], (0, 'CD1'): [(0, 'SC1')], (0, 'CD2'): [(0, 'SC1')]}
    extra = ()
    mappings = {'universal': {'martini22': {'GLY': (gly, _map_weights(gly), extra),
                                            'ILE': (ile, _map_weights(ile), extra),
                                            'LEU': (leu, _map_weights(leu), extra),}}}
    
    peptide = Molecule(force_field=FF_UNIVERSAL)
    aa = FF_UNIVERSAL.blocks['GLY'].to_molecule()
    peptide.merge_molecule(aa)

    aa = FF_UNIVERSAL.blocks['ILE'].to_molecule()
    peptide.merge_molecule(aa)

    aa = FF_UNIVERSAL.blocks['LEU'].to_molecule()
    peptide.merge_molecule(aa)

    peptide.add_edge(list(peptide.find_atoms(atomname='N', resid=1))[0],
                     list(peptide.find_atoms(atomname='C', resid=2))[0])
    peptide.add_edge(list(peptide.find_atoms(atomname='N', resid=2))[0],
                     list(peptide.find_atoms(atomname='C', resid=3))[0])

    for node in peptide:
        peptide.nodes[node]['atomid'] = node + 1
        peptide.nodes[node]['chain'] = ''

    cg = do_mapping(peptide, mappings, FF_MARTINI)

    expected = Molecule(force_field=FF_MARTINI)
    expected.add_nodes_from({1: {'atomname': 'BB',
                                 'atype': 'P5',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 1,
                                 'resid': 1,
                                 'resname': 'GLY'},
                             2: {'atomname': 'BB',
                                 'atype': 'P5',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 2,
                                 'resid': 2,
                                 'resname': 'ILE'},
                             3: {'atomname': 'SC1',
                                 'atype': 'AC1',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 2,
                                 'resid': 2,
                                 'resname': 'ILE'},
                             4: {'atomname': 'BB',
                                 'atype': 'P5',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 3,
                                 'resid': 3,
                                 'resname': 'LEU'},
                             5: {'atomname': 'SC1',
                                 'atype': 'AC1',
                                 'chain': '',
                                 'charge': 0.0,
                                 'charge_group': 3,
                                 'resid': 3,
                                 'resname': 'LEU'}}.items()
                            )
    expected.add_edges_from([(1, 2), (2, 3), (2, 4), (4, 5)])
    
    for node in expected:
        expected.nodes[node]['atomid'] = node + 1
    
    assert _equal_graphs(cg, expected)

if __name__ == '__main__':
    test_peptide()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that can perform a resolution transformation on a
molecule.
"""
from collections import defaultdict, Counter
from functools import partial
from itertools import product, combinations

import networkx as nx

from ..molecule import Molecule
from .processor import Processor
from ..utils import are_all_equal, format_atom_string
from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


class GraphMapping:
    # Attributes to be removed from the blocks.
    forbidden = ['charge_group']

    # TODO: Add __getitem__, __iter__, __len__, __contains__, keys, values and
    #       items methods to emulate a Mapping?

    # TODO: Different methods of initializing the mapping? It might be nice to
    #       also provide the option to provide two molecules (instead) of lists
    #       of blocks and a mapping of {node_idx: [node_idx, ...], ...}
    def __init__(self, blocks_from, blocks_to, mapping, weights=None, extra=()):
        """
        blocks_from and blocks_to are sequences of Blocks.
        Mapping is a dictionary of {(residx, atomname): [(residx, atomname), ...], ...}.
        residx in these cases is the index of the residue in blocks_from and
        blocks_to respectively.
        """
        if weights is None:
            weights = {}
        self.block_from = self._merge(blocks_from)
        self.block_to = self._merge(blocks_to)
        self.mapping = defaultdict(set)
        self.weights = defaultdict(dict)
        # Translate atomnames in mapping to node keys.
        for from_, to in mapping.items():
            res_from, name_from = from_
            # +1 here (and below), because the blockidxs in mapping are start
            # at 0, while resids start at 1.
            from_idxs = list(self.block_from.find_atoms(atomname=name_from, resid=res_from+1))
            for res_to, name_to in to:
                to_idxs = self.block_to.find_atoms(atomname=name_to, resid=res_to+1)
                for to_idx in to_idxs:
                    self.mapping[to_idx].update(from_idxs)
                    if from_idxs:
                        self.weights[to_idx][from_idxs[0]] = (
                            weights[(res_to, name_to)][(res_from, name_from)]
                        )

        self.mapping = dict(self.mapping)
        self.extra = extra

        # We can't do this in _merge, since we need the resids to translate the
        # mapping from (ambiguous) atomnames to (unique) graph keys. We do have
        # to get rid of them, otherwise they overwrite the resids of the graph
        # we're mapping (in do_mapping).
        self._purge_forbidden(self.block_from)
        self._purge_forbidden(self.block_to)

        # Since we merged blocks, there may be edges missing in both (between
        # the provided blocks). This is bad. We should add that info to the
        # mapping, somehow.

        # One option is to add from eachother, but this tends to introduce
        # issues.
        # Example:
        #   from_blocks: a0-b0-c0 a1-b1-c1
        #   to_blocks:   A0-B0 C1-A1 B2-C2
        #   Mapping: {(0, A): [(0, a)], (0, B): [(0, b)], (1, C): [(0, c)],
        #             (1, A): [(1, a)], (2, B): [(1, b)], (2, C): [(1, c)]}
        # There are edges missing in both from_blocks and to_blocks (c0-a1, and
        # B0-C1 and A1-B2; respectively). There's enough information in the
        # "other" block to create those, so that's what we do.
#        for to_idx, to_jdx in self.block_to.edges():
#            try:
#                self.block_from.add_edges_from(product(self.mapping[to_idx],
#                                                       self.mapping[to_jdx]))
#            except KeyError:
#                # Either to_idx or to_jdx don't actually contribute to the
#                # mapping
#                pass
#            # e.g. for edge B0-C1 this is:
#            # add_edges_from(product([(0, b)], [(0, c)])); which is the same as
#            # add_edges_from(((0, b), (0, c)))
        # Cache the reverse map for a while. Maybe that means it shouldn't be
        # a property...
#        reverse_map = self.reverse_mapping
#        # This loop does the same as the one above, but in the other direction.
#        for from_idx, from_jdx in self.block_from.edges():
#            try:
#                self.block_to.add_edges_from(product(reverse_map[from_idx],
#                                                     reverse_map[from_jdx]))
#            except KeyError:
#                # Either from_idx or from_jdx don't actually contribute to the
#                # mapping
#                pass

    @classmethod
    def _merge(cls, blocks):
        out = blocks[0].to_molecule()
        for block in blocks[1:]:
            out.merge_molecule(block)
        return out

    @classmethod
    def _purge_forbidden(cls, block):
        for n_idx in block:
            node = block.nodes[n_idx]
            for attr in cls.forbidden:
                if attr in node:
                    del node[attr]

    @property
    def reverse_mapping(self):
        reverse_mapping = defaultdict(set)
        for to_idx, from_idxs in self.mapping.items():
            for from_idx in from_idxs:
                reverse_mapping[from_idx].add(to_idx)
        reverse_mapping = dict(reverse_mapping)
        return reverse_mapping


def build_graph_mapping_collection(from_ff, to_ff, mappings):
    graph_mapping_collection = {}
    pair_mapping = mappings[from_ff.name][to_ff.name]
    for name in from_ff.blocks.keys():
        if name in to_ff.blocks and name in pair_mapping:
            mapping, weights, extra = pair_mapping[name]
            graph_mapping_collection[name] = GraphMapping(
                [from_ff.blocks[name], ],
                [to_ff.blocks[name], ],
                mapping,
                weights,
                extra,
            )
    return graph_mapping_collection


# We can't inherit from nx.isomorphism.GraphMatcher to override
# `semantic_feasibility`. That implementation will clobber this one's method.
class MappingGraphMatcher(nx.isomorphism.isomorphvf2.GraphMatcher):
    def __init__(self, *args, edge_match=None, node_match=None, **kwargs):
        self.edge_match = edge_match
        self.node_match = node_match
        super().__init__(*args, **kwargs)
        self.G1_adj = self.G1.adj

    def semantic_feasibility(self, G1_node, G2_node):
        """
        Returns True if mapping G1_node to G2_node is semantically feasible.
        Adapted from networkx.algorithms.isomorphism.vf2userfunc._semantic_feasibility.
        """
        # Make sure the nodes match
        if self.node_match is not None:
            nm = self.node_match(self.G1.nodes[G1_node], self.G2.nodes[G2_node])
            if not nm:
                return False

        # Make sure the edges match
        if self.edge_match is not None:

            # Cached lookups
            core_1 = self.core_1
            edge_match = self.edge_match

            for neighbor in self.G1_adj[G1_node]:
                # G1_node is not in core_1, so we must handle R_self separately
                if neighbor == G1_node:
                    if not edge_match(G1_node, G1_node, G2_node, G2_node):
                        return False
                elif neighbor in core_1:
                    if not edge_match(G1_node, neighbor, G2_node, core_1[neighbor]):
                        return False
            # syntactic check has already verified that neighbors are symmetric
        return True


def edge_matcher(graph1, graph2, node11, node12, node21, node22):
    """
    Checks whether the resids for node11 and node12 in graph1 are the same, and
    whether that's also true for node21 and node22 in graph2.
    """
    node11 = graph1.nodes[node11]
    node12 = graph1.nodes[node12]
    node21 = graph2.nodes[node21]
    node22 = graph2.nodes[node22]
    return (node11.get('resid') == node12.get('resid')) ==\
           (node21.get('resid') == node22.get('resid'))


def do_mapping(molecule, mappings, to_ff, attribute_keep=()):
    """
    Creates a new :class:`~vermouth.molecule.Molecule` in force field `to_ff`
    from `molecule`, based on `mappings`. It does this by doing a subgraph
    isomorphism of all blocks in `mappings` and `molecule`. Will issue warnings
    if there's atoms not contibuting to the new molecule, or if there's
    overlapping blocks.
    Node attributes in the new molecule will come from the blocks constructing
    it, except for those in `attribute_keep`, which lists the attributes that
    will be kept from `molecule`.

    Parameters
    ----------
    molecule: :class:`~vermouth.molecule.Molecule`
        The molecule to transform.
    mappings: dict[str, dict[str, dict[str, tuple]]]
        ``{ff_name: {ff_name: {block_name: (mapping, weights, extra)}}}``
        A collection of mappings, as returned by e.g.
        :func:`~vermouth.map_input.read_mapping_directory`.
    to_ff: :class:`~vermouth.forcefield.ForceField`
        The force field to transform to.
    attribute_keep: :class:`~collections.abc.Iterable`
        The attributes to keep from `molecule`

    Returns
    -------
    :class:`~vermouth.molecule.Molecule`
        A new molecule, created by transforming `molecule` to `to_ff` according
        to `mappings`.
    """
    # Transfering the meta meybe should be a copy, or a deep copy...
    # If it breaks we look at this line.
    graph_out = Molecule(force_field=to_ff, meta=molecule.meta)
    # We want to keep the 'chain' property from the original molecule.
    attribute_keep = ['chain'] + list(attribute_keep)
    pair_mapping = build_graph_mapping_collection(molecule.force_field, to_ff, mappings)
    all_matches = []
    for resname, mapping in pair_mapping.items():
        # TODO: add PTMs as a matching criterion here.
        # Make sure the atomname and resname match
        node_match = nx.isomorphism.categorical_node_match(['atomname', 'resname'], ['', ''])
        # And make sure that we don't accidentally cross a residue boundary,
        # unless that's allowed by the mapping.
        edge_match = partial(edge_matcher, molecule, mapping.block_from)
        # We're going to find *every* way block fits on molecule.
        graphmatcher = MappingGraphMatcher(molecule, mapping.block_from,
                                           node_match=node_match, edge_match=edge_match)
        matches = graphmatcher.subgraph_isomorphisms_iter()
        for match in matches:
            all_matches.append((match, resname, mapping))
    mol_to_out = defaultdict(list)
    blocks_per_atom = Counter()
    # Sort by lowest node key per residue. We need to do this, since
    # merge_molecule creates new resid's in order.
    for match, name, mapping in sorted(all_matches, key=lambda x: min(x[0].keys())):
        blocks_per_atom.update(match.keys())
        if graph_out.nrexcl is None:
            graph_out.nrexcl = mapping.block_to.nrexcl
        try:
            # merge_molecule will return a dict mapping the node keys of the
            # added block to the ones in graph_out
            block_to_out = graph_out.merge_molecule(mapping.block_to)
        except ValueError:
            # This probably means the nrexcl of the block is different from the
            # others. This means the user messed up their data. Or there are
            # different forcefields in the same forcefield folder...
            LOGGER.exception('Residue {} is not compatible with the others',
                             name, type='inconsistent-data')
            raise
        block_to_mol = {v: k for k, v in match.items()}
        for to_idx, from_idxs in mapping.mapping.items():
            # Some bookkeeping with indices.
            out_idx = block_to_out[to_idx]
            mol_idxs = [block_to_mol[from_idx] for from_idx in from_idxs]
            for mol_idx in mol_idxs:
                mol_to_out[mol_idx].append(out_idx)

            # Keep track of what bead comes from where
            subgraph = molecule.subgraph(mol_idxs)
            graph_out.nodes[out_idx]['graph'] = subgraph
            weights = {block_to_mol[from_idx]: mapping.weights[to_idx][from_idx]
                       for from_idx in from_idxs}
            graph_out.nodes[out_idx]['mapping_weights'] = weights
            # We drop the node keys, since those are not super relevant. We are
            # just interested in values of the node attributes, and whether
            # they're all equal.
            attrs = {name: list(nx.get_node_attributes(subgraph, name).values())
                     for name in attribute_keep}
            for attr, vals in attrs.items():
                if not are_all_equal(vals):
                    LOGGER.warning('The attribute {} for atom {} is going to'
                                   ' be garbage.', name, format_atom_string(graph_out.nodes[out_idx]),
                                   type='inconsistent-data')
                if vals:
                    graph_out.nodes[out_idx][attr] = vals[0]
                else:
                    # No nodes hat the attribute `name`. And
                    # nx.get_ndoe_attributes doesn't take a default.
                    graph_out.nodes[out_idx][attr] = None
    mol_to_out = dict(mol_to_out)
    # We need to add edges between residues. Within residues comes from the
    # blocks.
    # TODO: backmapping needs some magic here.
    for match1, match2 in combinations(all_matches, 2):
        match1 = match1[0]
        match2 = match2[0]
        edges = molecule.edges_between(match1.keys(), match2.keys())
        for mol_idx, mol_jdx in edges:
            out_idxs = mol_to_out[mol_idx]
            out_jdxs = mol_to_out[mol_jdx]
            for out_idx, out_jdx in product(out_idxs, out_jdxs):
                if out_idx != out_jdx:
                    graph_out.add_edge(out_idx, out_jdx)
        shared_atoms = set(match1.keys()) & set(match2.keys())
        shared_out_atoms = [mol_to_out[mol_idx] for mol_idx in shared_atoms]
        for in_atom, out_atoms in zip(shared_atoms, shared_out_atoms):
            if len(out_atoms) < 2:
                LOGGER.critical('The atom {} is shared between blocks, but'
                                ' only mapped once to {}?',
                                format_atom_string(molecule.nodes[in_atom]),
                                [format_atom_string(graph_out.nodes[idx])
                                 for idx in out_atoms],
                                type='inconsistent-data')
                raise ValueError("This atom is shared between blocks, but only"
                                 " mapped once?")
            for out_idx, out_jdx in combinations(out_atoms, 2):
                if out_idx != out_jdx:
                    graph_out.add_edge(out_idx, out_jdx)
        if shared_atoms:
            LOGGER.warning("You have the following atoms that are shared"
                           " between blocks. This may mean you have too many"
                           " particles in your output and/or erroneous bonds."
                           " {}. They've end up in the following output atoms:"
                           " {}.",
                           [format_atom_string(molecule.nodes[idx], atomid=idx) for idx in shared_atoms],
                           [format_atom_string(graph_out.nodes[idx], atomid=idx) for idxs in shared_out_atoms for idx in idxs],
                           type='inconsistent-data')

    # Sanity check the results
    if any(v > 1 for v in blocks_per_atom.values()):
        LOGGER.warning('These atoms are covered by multiple blocks. This is a '
                       'bad idea: {}', {format_atom_string(molecule.nodes[k], atomid=k): v
                                        for k, v in blocks_per_atom.items() if v > 1},
                       type='inconsistent-data')
    uncovered_atoms = set(molecule.nodes.keys()) - set(mol_to_out.keys())
    if uncovered_atoms:
        uncovered_hydrogens = {idx for idx in uncovered_atoms
                               if molecule.nodes[idx].get('element', '') == 'H'}
        if uncovered_hydrogens:
            # Maybe this should be info?
            LOGGER.debug('These hydrogen atoms are not covered by a mapping.'
                         ' This is not the best idea. {}',
                         [format_atom_string(molecule.nodes[idx])
                          for idx in uncovered_hydrogens],
                         type='unmapped-atom'
                        )
        other_uncovered = uncovered_atoms - uncovered_hydrogens
        if other_uncovered:
            LOGGER.warning("These atoms are not covered by a mapping. Either"
                           " your mappings don't describe all atoms (bad idea),"
                           " or, there's no mapping available for all residues."
                           " {}",
                           [format_atom_string(molecule.nodes[idx])
                            for idx in other_uncovered],
                           type='unmapped-atom')
    return graph_out


class DoMapping(Processor):
    def __init__(self, mappings, to_ff, delete_unknown=False, attribute_keep=()):
        self.mappings = mappings
        self.to_ff = to_ff
        self.delete_unknown = delete_unknown
        self.attribute_keep = attribute_keep
        super().__init__()

    def run_molecule(self, molecule):
        return do_mapping(
            molecule,
            mappings=self.mappings,
            to_ff=self.to_ff,
            attribute_keep=self.attribute_keep
        )

    def run_system(self, system):
        mols = []
        for molecule in system.molecules:
            try:
                new_molecule = self.run_molecule(molecule)
            except KeyError as err:
                if not self.delete_unknown:
                    raise err
                else:
                    raise
                    # TODO: raise a loud warning here
            else:
                if new_molecule:
                    mols.append(new_molecule)
        system.molecules = mols
        system.force_field = self.to_ff
